/**
 * Model-based Level Set (MLS) Algorithm
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2003-2005 Audrey H. Zhuang and Daniel J. Valentino
 *
 * Please read LICENSE.TXT for the license covering this software
 *
 * For more information, please contact the authors at:
 * haihongz@seas.ucla.edu
 * dvalentino@mednet.ucla.edu
 */
package mlsSkullStrip;

import java.io.File;
import java.util.Vector;

/**
 * Application to analyze user's command line.
 * 
 * @author Audrey H. Zhuang and Daniel J. Valentino
 * @version 2 September 2005
 */ 
public class CommandLineParser
{

    /** An input file. */
    private File _inputFile = null;

    /** An output mask file. */
    private File _outputMaskFile = null;

    /** An output brainOnly file.*/
    private File _outputBrainOnlyFile = null;

    private double _velocity = 0.05 * 255 * 255;

    private boolean _isDisplayingIntermediateResultsEnabled = false;

    /** Whether to fill holes generated by skull stripping algorithm. */
    private boolean _isFillingHolesEnabled = true;

    /** Threshold selector. */
    private double _thresholdSelector = 0.5;

    /** The way where the head locates: 0-north, 1-east, 2-south, 3-west. */
    private int _heading = 0;

    /**
     * Constructs UserCommandAnalyzer.
     * 
     * @param args user's command stored in a String array.
     */
    public CommandLineParser(String fileNames[])
    {
	_parse(fileNames);
    }

    /**
     * Parse user's command line.
     *
     * @param args user's command stored in a String array.
     */
    private void _parse(String args[])
    {
      // The directory shared by the files
      String d = null;
       for(int i = 0; i < args.length; i++){

	  if( args[i].startsWith("-i")) {
	      i++;
	      _inputFile = new File(args[i]);
	  }
	  else if( args[i].startsWith("-m")){
	      i++;
	      _outputMaskFile = new File(args[i]);
	  } 
	  else if( args[i].startsWith("-b")){
	      i++;
	      _outputBrainOnlyFile = new File(args[i]);
	  } 
	  else if(args[i].startsWith("-v")){
	      while(!(i+1 >= args.length || args[i+1].startsWith("-"))){
		  i++;
		  _velocity = Double.valueOf(args[i]).doubleValue() *255 * 255;
	      }
	  }
	  else if(args[i].startsWith("-t")){
	      while(!(i+1 >= args.length || args[i+1].startsWith("-"))){
		  i++;
		  _thresholdSelector = Double.valueOf(args[i]).doubleValue();
	      }    
	  }
	  else if (args[i].startsWith("-p")){
	      i++;
	      _heading = Integer.valueOf(args[i]).intValue();
	  }
       }
    }


    /**
     * Get heading direction.
     *
     * @return In integer indicating the heading direction.  
     *         0-north, 1-east, 2-south, 3-west.
     */
    public int getHeadingDirection()
    {
	return _heading;
    }

    /**
     * Gets the input file.
     *
     * @return The input file.
     */
    public File getInputFile()
    {
	return _inputFile;
    }


    /**
     * Gets the output mask file.
     *
     * @return The output mask file.
     */
    public File getOutputMaskFile()
    {
	return _outputMaskFile;
    }

    /**
     * Gets the output brain only file.
     *
     * @return The output brain only file.
     */
    public File getOutputBrainOnlyFile()
    {
	return _outputBrainOnlyFile;
    }


    /**
     * Get the threshold selector which is a float number between 0 and 1.
     *
     * @return a float number.
     */
    public double getThresholdSelector()
    {
	return _thresholdSelector;
    }

    public double getVelocity()
    {
	return _velocity;
    }

    public boolean isDisplayingIntermediateResultsEnabled()
    {
	return _isDisplayingIntermediateResultsEnabled;
    }

    public boolean isFillingHolesEnabled()
    {
	return _isFillingHolesEnabled;
    }
}
