/**
 * Model-based Level Set (MLS) Algorithm
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2003-2005 Audrey H. Zhuang and Daniel J. Valentino
 *
 * Please read LICENSE.TXT for the license covering this software
 *
 * For more information, please contact the authors at:
 * haihongz@seas.ucla.edu
 * dvalentino@mednet.ucla.edu
 */
package mlsSkullStrip;

import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster; 
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import javax.imageio.ImageReader;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.BevelBorder;

/**
 * Class to read images from the file names provided on the command 
 * line as arguments and execute skull-stripping task on them.
 *
 * @author Audrey H. Zhuang and Daniel J. Valentino
 * @version 2 September 2005
 */
public class DataManager
{

  /** A vector contains a list of objects of SkullStrippers. */
  private Vector _skullStrippers = new Vector(1, 1);

  /** Object of AlphaViewport5. */
  private Vector _alphaViewports = new Vector(1, 1);

  /** Maximum iterations. */
  private int MAX_ITER = 255*16; 

  /** Object of ImageReader. */
  private ImageReader _imageReader;
 
  /** Format type of the given file. */
  private String _format = "analyze";

  /** Format names supported by volume constructor.*/
  public static final String[] FORMAT_TYPES = {"analyze", "minc", "dicom", "ge", 
					  "ucf", "dvazio"};
  /** The function phi. */
  private double[][] _phi = null;

  /** The slice index with which the skull stripping starts with. */
    private int _initialIndex = 0;


  /** 
   * The intensity minimum below which lies 2% of the cumulative
   * histogram. 
   */ 
  private double _intens2 = 0;

  /** It is the intensity maxmum below which lies 98% of the cumulative
   * histogram.
   */
  private double _intens98 = 0;
  private double _intens10 = 0;
  private double _intens40 = 0;   

  /** Interslice distance (/pixel size). */
  private float INTER_SLICE_DIST = 3;

  /** The orientation of 2D slices resampled from the data volume and stored 
   * in the skullStrippers. */
  private int _resamplingOrient = 0;

  /** Orientation of the 2D image in the loaded file.*/
  private int _originalOrient = 0;

  /** Bits per pixel. */
  private int _bitsPP = 16;

  /** The way where the head locates: 0-north, 1-east, 2-south, 3-west. */
  private int _heading = 0;

  /** Whether to fill holes generated by skull stripping algorithm. */
  private boolean _isFillingHolesEnabled = false;

  /** Cut off boundary value of phi. */
  private double BOUNDARY_VALUE = -0.5;

  public DataManager(){ }  

  /**
   * Gets the skullStripper at the given index.
   *
   * @param index The index of the skullStripper.
   */
  public SkullStripper getSkullStripper(int index)
  {
    return (SkullStripper) _skullStrippers.elementAt(index);
  }

  /**
   * Gets the slice index of the image with which the skull stripping starts 
   * with.
   *
   * @return An integer number. 
   */
  public int getInitialIndex()
  {
      return _initialIndex;
  }

  /**
   * Fill the holes in the brain masks.
   */
  public void fillHoles()
  {
    for(int i = 0; i < _skullStrippers.size(); i++){
	SkullStripper skullStripper = (SkullStripper)_skullStrippers
	    .elementAt(i);
      byte[][] outputArray1 = skullStripper.getOutputArray();
      Filler filler = new Filler(outputArray1);
      byte[][] outputArray2 = filler.getFilledArray();
      skullStripper.setOutputArray(outputArray2);
    }
  }

  /**
   * Fill the holes in a brain mask stored in a given skullStripper object.
   */
  public void fillHoles(SkullStripper skullStripper)
  {
    byte[][] outputArray1 = skullStripper.getOutputArray();
    Filler filler = new Filler(outputArray1);
    byte[][] outputArray2 = filler.getFilledArray();
    skullStripper.setOutputArray(outputArray2);
  }


  /**
   * Initialize a zero level set circle in a 2D slice resampled from the 
   * input file.
   */
  public void initializeZeroLS()
  {
    ZeroLSInitializer initializer = new ZeroLSInitializer(_imageReader,
							  _skullStrippers,
							  _intens2,
							  _intens98,
							  _resamplingOrient);
    double[] params = initializer.getInitialParameters();
    double x = params[0];
    double y = params[1];
    _initialIndex = (int) params[2];
    double r = params[3];

    // Get an SkullStripper object at the given index
    SkullStripper ac5 = getSkullStripper(_initialIndex);

    // Set initialization parameters and initialize
    ac5.setInitialParameters(x, y, r);    
    _phi = ac5.initializePhi();
  }

  /**
   * Load the file to segment.
   *
   * @param file File to segment. Assume it is not null.
   * @param parser Command line parser which provides user specified 
   *               parameters.
   */
  public void loadFile(File file, CommandLineParser parser)
      throws IOException
  {

    // Set the value for velocity and thresholdSelector
    double velocity = 0;
    double thresholdSelector = 0;
    if(parser != null) {
	velocity = parser.getVelocity();
	thresholdSelector = parser.getThresholdSelector();
	_heading = parser.getHeadingDirection();
      if(_heading > 3) _heading = 0;
      _isFillingHolesEnabled = parser.isFillingHolesEnabled();
    }
    else {

	// Set the default velocity and thresholdSelector;
	velocity = 0.05 * 255 * 255;
	thresholdSelector = 0.5;
    }
    
    FileReader fileReader = new FileReader(file);
    _imageReader = fileReader.getImageReader();

    try{
	// Get original and resampling orient
	MetadataParser metadataParser = new MetadataParser(_imageReader);
	_format = metadataParser.getFormatName();
	_originalOrient = metadataParser.getOrient();
	DataVolume dv = new DataVolume(_imageReader);
	_resamplingOrient = _getResamplingOrient(metadataParser, dv);
	_bitsPP = metadataParser.getBitsPP();

	// Resample data volume into 2D slices
	BufferedImage[] imgs = null;
	if(_resamplingOrient == _originalOrient){
	  int size = _imageReader.getNumImages(true);
	  imgs = new BufferedImage[size];
	  for(int i = 0; i < imgs.length; i++){
	      imgs[i] = _imageReader.read(i);
	  }
	}
	else {
	  DataVolumeResampler dvResampler = new DataVolumeResampler(dv);
	  int IMAGE_TYPE = BufferedImage.TYPE_USHORT_GRAY;
	  if(_bitsPP <= 8) IMAGE_TYPE = BufferedImage.TYPE_BYTE_GRAY;
	  else IMAGE_TYPE = BufferedImage.TYPE_USHORT_GRAY;
	  int size = dv.getDepth(_resamplingOrient);
	  imgs = new BufferedImage[size];
	  for(int i = 0; i < imgs.length; i++){
	    imgs[i] = dvResampler.createGrayImage(IMAGE_TYPE, i, 
						  _resamplingOrient);
	  }
	}
	float xyDim = _getXYDim(metadataParser);
	double age = metadataParser.getAge();
	INTER_SLICE_DIST = _getInterSliceDist(metadataParser, 
					      _resamplingOrient);

	if(imgs == null){
	  String msg = "Cannot get image data from the input file!";
	  throw new IOException(msg);
	} 

	// Create skullStrippers for 2D slices
	BufferedImage img = null;
	for(int i = 0; i < imgs.length; i++){
	  img = imgs[i];
	  SkullStripper ac = new SkullStripper(img, velocity,
						 thresholdSelector);
	  ac.setBitsPP( _bitsPP );
	  ac.calculateProbingDistance(xyDim, age);
	  _skullStrippers.add(ac);
	}

	// Calculate histogram 
	_histogram(_skullStrippers);

	// Set intens2, intens10, intens98 to skullStrippers
	SkullStripper ac = null;
	for(int i = 0; i < _skullStrippers.size(); i++){
	  ac = (SkullStripper)_skullStrippers.elementAt(i);
	  ac.setThresholdIntensities(_intens2, _intens10, _intens40,_intens98);

	}
      }catch(Exception e){
	  e.printStackTrace();
      } 
  }

  /**
   * Save skull-stripped images.
   * 
   * @param fileName The name of the file to save.
   */
  public void saveBrainOnlyFile(String fileName)
  {
    FileWriter writer = new FileWriter();
    BufferedImage[] images = new BufferedImage[_skullStrippers.size()];

    // Get skull-stripped 2D slices from skullStrippers
    for(int i = 0; i < images.length; i++ ){
	images[i] = ((SkullStripper)_skullStrippers.elementAt(i))
	    .getOutputImage();
    }

    if(_originalOrient == _resamplingOrient){
	writer.saveFile(fileName, images, _imageReader);
    }

    // If the resamplingOrient is different from the originalOrient, reslice
    // the data volume according to originalOrient
    else {
      MetadataParser mp = new MetadataParser(_imageReader);
      String format = mp.getFormatName();
      DataVolume dv = new DataVolume(images, _resamplingOrient, _bitsPP, 
				     format);
      DataVolumeResampler dvResampler = new DataVolumeResampler(dv);
      int originalDepth = dv.getDepth(_originalOrient);
      int IMAGE_TYPE = BufferedImage.TYPE_USHORT_GRAY;
      if(_bitsPP <= 8) IMAGE_TYPE = BufferedImage.TYPE_BYTE_GRAY;
      BufferedImage[] outputImages = new BufferedImage[originalDepth];
      for(int i = 0; i < outputImages.length; i++ ){
	  outputImages[i] = dvResampler.createGrayImage(IMAGE_TYPE, i, 
							_originalOrient);
      }
      writer.saveFile(fileName, outputImages, _imageReader);
    }
  } 

  /**
   * Save binary images in which 0 represents background and non-zero 
   * represents brain tissues. The metadata to save with the skull-stripped 
   * results are obtained by revising the metadata of the original image.
   *
   * @param baseName The name of the file to save, and the extension of the 
   *                 file name, e.g., .mnc, .hdr, .img etc., will be added
   *                 by FileWriter class.
   */
  public void saveMaskFile(String fileName)
  {

    // Save the brain masks as Byte images
    BufferedImage[] images = new BufferedImage[_skullStrippers.size()];
    FileWriter writer = new FileWriter();
    for(int i = 0; i < images.length; i++ ){
      images[i] = ((SkullStripper)_skullStrippers.elementAt(i))
	  .getOutputMaskImage();
    }
    int intensityOfMask = ((SkullStripper)_skullStrippers.elementAt(0))
	.getIntensityOfMask();

    if(_originalOrient == _resamplingOrient){
      writer.saveMaskFile(fileName, images, intensityOfMask, _imageReader);
    }

    // If the resamplingOrient is different from the originalOrient, reslice
    // the data volume according to originalOrient
    else {
      MetadataParser mp = new MetadataParser(_imageReader);
      String format = mp.getFormatName();
      DataVolume dv = new DataVolume(images, _resamplingOrient, _bitsPP, 
				     format);
      DataVolumeResampler dvResampler = new DataVolumeResampler(dv);
      int originalDepth = dv.getDepth(_originalOrient);
      int IMAGE_TYPE = BufferedImage.TYPE_USHORT_GRAY;
      if(_bitsPP <= 8) IMAGE_TYPE = BufferedImage.TYPE_BYTE_GRAY;
      BufferedImage[] outputImages = new BufferedImage[originalDepth];
      for(int i = 0; i < outputImages.length; i++ ){
	  outputImages[i] = dvResampler.createGrayImage(IMAGE_TYPE, i, 
							_originalOrient);
      }
      writer.saveMaskFile(fileName, outputImages,intensityOfMask,_imageReader);
    }
  }

  /**
   * Perform skull stripping.
   */
  public void doSkullStripping()
  {
      initializeZeroLS();
      setStrippingEnabled(_initialIndex);
  }

  /**
   * Sets skull stripping enabled and do the skull-stripping.
   *
   * @param index The index of the 2D slice to start skull-stripping with.
   */
  public void setStrippingEnabled(int index )
  {
    if(_skullStrippers == null)return;
    if(_alphaViewports == null) return;
    int SHRINK_DIST = (int) Math.floor(Math.abs(INTER_SLICE_DIST) + 0.5);
    int lastMaskArea = 0;
    int maskArea = 0;
    int contourLength = 0;

    // The slices, that are around mid-sagittal slices, are estimated to be 
    // within SLICE_1 and SLICE_2
    double temp = _skullStrippers.size();
    int SLICE_1 = (int) (temp * 0.45);
    int SLICE_2 = (int) (temp * 0.55);

    // Define the length of a small contour
    int SMALL_CIRCLE = 100;  
    double[][] phiTemp = null;
    boolean isPhiAcceptable = true;
    double velocity = 0;
    double tempV = 0;

    // Notify user that DeSkulling has started
    System.out.print( "Deskulling " );
    System.out.print( "." );

    // Do skull-stripping on the slice with the "index"
    SkullStripper ac = (SkullStripper)_skullStrippers.elementAt(index);
    _phi = ac.skullStrip(MAX_ITER, _phi);
    if(_isFillingHolesEnabled) fillHoles(ac);
    BufferedImage img = ac.getZeroLevelSetImage();

    // Keep the _phi for slides before "index"
    double[][] phiBackword = _phi;

    // Skull-strip slices after "index"
    for(int i = (index+1); i < _skullStrippers.size(); i++ ){
      System.out.print( "." );

      ac = (SkullStripper)_skullStrippers.elementAt(i);
      phiTemp = _phi;

      // Shrink the phi zero contour before applying it as an initial contour 
      // Using reiniailization method 
      _phi = ac.shrinkPhiZero(_phi, SHRINK_DIST);
      _phi = ac.skullStrip(MAX_ITER, _phi);

      // Do not check acceptance if the images are around mid-sagittal slices
      // where the CSF change is big and segmentation between slices could be
      // very different
      if(_resamplingOrient == 2 && i > SLICE_1 && i < SLICE_2) 
	  isPhiAcceptable = true;
      else {
	  // Check if the skull-stripped result is acceptable
	  isPhiAcceptable = _isPhiAcceptable(_phi, ac, phiTemp);
      }
      if( !isPhiAcceptable){

	// Increase velocity by 5 times
	velocity = ac.getVelocity();
	tempV = velocity * 5;
	ac.setVelocity(tempV);
	ac.setStrippingEnabled(true);
	_phi = ac.shrinkPhiZero( phiTemp, SHRINK_DIST + 10);
	_phi = ac.skullStrip(MAX_ITER, _phi);
	
	// Check if the skull-stripped result is  acceptable
	isPhiAcceptable = _isPhiAcceptable(_phi, ac, phiTemp);

	if( !isPhiAcceptable){

	  // Increase velocity by 10 times 
	  tempV = velocity * 10;
	  ac.setVelocity(tempV);	
	  ac.setStrippingEnabled(true);
	  _phi = ac.shrinkPhiZero( phiTemp, SHRINK_DIST + 5);
	  _phi = ac.skullStrip(MAX_ITER, _phi);
	  
	  // Check if the skull-stripped result is acceptable
	  isPhiAcceptable = _isPhiAcceptable(_phi, ac, phiTemp);	  
	}
	if( !isPhiAcceptable){
	  contourLength = ac.calculateContourLength(phiTemp);
	  if( contourLength > SMALL_CIRCLE ){

	      // Use mask in last slide for current slide
	      _phi = ac.skullStrip(0, phiTemp);
	  }
	  else {

	      // Shrink the circle to zero
	      while( ac.calculateMaskArea(_phi) > 0){
		  int shrinkDistTemp = 4;
		  _phi = ac.shrinkPhiZero(_phi, shrinkDistTemp);
	      }
	      _phi = ac.skullStrip(0, _phi);
	  }
	} 
      }
      if(_isFillingHolesEnabled) fillHoles(ac);
    }


    // Skull-strip slides before "index"
    _phi = phiBackword;
     ac = (SkullStripper)_skullStrippers.elementAt(index);
    for(int i = (index-1); i >= 0; i--){
      System.out.print( "." );
      ac = (SkullStripper)_skullStrippers.elementAt(i);
      phiTemp = _phi;

      // Shrink the phi zero contour before applying it as an initial contour 
      _phi = ac.shrinkPhiZero(_phi, SHRINK_DIST);
      _phi = ac.skullStrip(MAX_ITER, _phi);

      // Do not check acceptance if the images are around mid-sagittal slices
      // where the CSF change is big and segmentation between slices could be
      // very different
      if(_resamplingOrient == 2 && i > SLICE_1 && i < SLICE_2) 
	  isPhiAcceptable = true;
      else {
	  // Check if _phi is acceptable
	  isPhiAcceptable = _isPhiAcceptable(_phi, ac, phiTemp);
      }
      if( !isPhiAcceptable){
	  velocity = ac.getVelocity();
 
	  // Increase the weight of curvature by 5 times 
	  tempV = velocity * 5;
	  ac.setVelocity(tempV);
	  ac.setStrippingEnabled(true);
	  _phi = ac.shrinkPhiZero( phiTemp, SHRINK_DIST + 10);
	  _phi = ac.skullStrip(MAX_ITER, _phi);

	  // Check if _phi is acceptable
	  isPhiAcceptable = _isPhiAcceptable(_phi, ac, phiTemp);
	  if( !isPhiAcceptable){
 
	    // Increase the weight of curvature by 10 times 
	    tempV = velocity * 10;
	    ac.setVelocity(tempV);
	    ac.setStrippingEnabled(true);
	    _phi = ac.shrinkPhiZero( phiTemp, SHRINK_DIST + 5);
	    _phi = ac.skullStrip(MAX_ITER, _phi);

	    // Check if _phi is acceptable
	    isPhiAcceptable = _isPhiAcceptable(_phi, ac, phiTemp);
	  }
	  if( !isPhiAcceptable){
	      contourLength = ac.calculateContourLength(phiTemp);
	      if( contourLength > SMALL_CIRCLE ){
		  // Use mask in last slide for current slide
		  _phi = ac.skullStrip(0, phiTemp);
	      }
	      else {
		  // Shrink the circle to zero
		  while( ac.calculateMaskArea(_phi) > 0){
		      int shrinkDistTemp = 4;
		      _phi = ac.shrinkPhiZero(_phi, shrinkDistTemp);
		  }
		  _phi = ac.skullStrip(0, _phi);
	      }
	  } 
      }
      if(_isFillingHolesEnabled) fillHoles(ac);
    } 
  }
  static {
      List formats = Arrays.asList(FORMAT_TYPES);
      if( ! formats.contains("analyze"))  
	  throw new InternalError("Doesn't support analyze file format!");
      if( ! formats.contains("minc"))  
	  throw new InternalError("Doesn't support minc file format!");
      if( ! formats.contains("ge"))  
	  throw new InternalError("Doesn't support ge file format!");
      if( ! formats.contains("dicom"))  
	  throw new InternalError("Doesn't support dicom file format!");
      if( ! formats.contains("ucf"))  
	  throw new InternalError("Doesn't support ucf file format!");
      if( ! formats.contains("dvazio"))  
	  throw new InternalError("Doesn't support dvazio file format!");
  }
    
  /**
   * Evaluate the skull-stripping algorithm.
   *
   * @param phi an array containing the family of the level sets.
   * @param labelBImg a buffered image which serves as the golden standard. 
   * @return An one element array of double values. The element is the 
   *         Jaccard coefficient of the given phi2 compared to the given phi1. 
   */
  private double[] _evaluate(double[][] phi1, double[][] phi2)
  {

    // True positive is successfully segmented pixels
    double TP = 0;

    // Pixels in the segmentation result that do not belong to the golden 
    // standard: false positive
    double FP = 0;

    // pixels of the golden standard but do not belong to the segmentation
    // results: false negative
    double FN = 0;
    
    double var1 = 0;
    double var2 = 0;
    for(int x = 0; x < phi1.length; x++){
      for(int y = 0; y < phi1[0].length; y++){
	var1 = phi1[x][y];
	var2 = phi2[x][y];

	// True positive
	if(var1 >= BOUNDARY_VALUE && var2 >= BOUNDARY_VALUE) TP += 1;
	
	// False positive
	if(var1 >= BOUNDARY_VALUE && var2 < BOUNDARY_VALUE) FP += 1;
	
	// False negative
	if(var1 < BOUNDARY_VALUE && var2 >= BOUNDARY_VALUE) FN += 1;
      }
    }    
    double jaccard = TP / (TP + FP + FN);
    double[] ans = new double[2];
    ans[0] = jaccard;
    return ans;
  }

  /**
   * Get resampling orient that is determined by the orientation giving the 
   * shortest depth. For example, if resampling in x direction gives the 
   * shortest depth, then the resampling orientation is assigned as axial.  
   * 
   * @param metadataParser An parser for parsing image file's metadata.
   * @param dv An dataVolume object storing the data.
   * @return An int value indicating the orientation: 0-axial, 1-coronal, 
   *         2-sagittal.
   */
  private int _getResamplingOrient(MetadataParser metadataParser,DataVolume dv)
  {
    int resamplingOrient = 0;


    if(dv.getDepth(0)<dv.getDepth(1) && dv.getDepth(0)<dv.getDepth(2)){
	resamplingOrient = 0;
    }
    else if(dv.getDepth(1)<dv.getDepth(2)){
	resamplingOrient = 1;
    }
    else resamplingOrient = 2;

    return resamplingOrient;
  }

  /**
   * Get the interslice distance. 
   *
   * @param metadataParser An parser for parsing image file's metadata.
   * @param resamplingOrient The resampling orientation.
   * @return A float number indicating the inter-slice distance (unit: mm).
   */
  private float _getInterSliceDist(MetadataParser metadataParser, 
				   int resamplingOrient)
  {
    float interSliceDist = 1.0f;
    float[] pixDims = metadataParser.getPixDims();

    if(resamplingOrient == 0) interSliceDist = pixDims[2];
    else if(resamplingOrient == 1) interSliceDist = pixDims[1];
    else interSliceDist = pixDims[0];
    return interSliceDist;
  }

  /**
   * Get the 2D resolusions. The pixDims contains 3D resolusions. The 2D
   * will be obtained by comparing the three elements of pixDims array,
   * and by assuming that the pixels are square, that is, the 
   * resolusions are the same in x and y directions. So if two elements 
   * in pixDims are the same, it is considered the 2D resolusions,
   * 
   * @param metadataParser An parser for parsing image file's metadata.
   * @return A float number indicating the resolution of the pixels (unit: mm).
   */
    private float _getXYDim(MetadataParser metadataParser)
    {
	float xyDim = 1.0f;
	float[] pixDims = metadataParser.getPixDims();

	// Get the 2D resolusions. The pixDims contains 3D resolusions. The 2D
	// will be obtained by comparing the three elements of pixDims array,
	// and by assuming that the pixels are square, that is, the 
	// resolusions are the same in x and y directions. So if two elements 
	// in pixDims are the same, it is considered the 2D resolusions,
	if( Math.abs(pixDims[0]-pixDims[1]) < (pixDims[0] * 0.1))
	    xyDim = pixDims[0];
	else if (Math.abs(pixDims[0]-pixDims[2]) < (pixDims[0] * 0.1))
	    xyDim = pixDims[0];
	else if (Math.abs(pixDims[1]-pixDims[2]) < (pixDims[1] * 0.1))
	    xyDim = pixDims[1];
	else xyDim = 1.0f; 
	return xyDim;
    }

  /**
   * Calculate the histogram of an image and initialize the class members of
   * _intens2 and _intens98.
   */
  private void _histogram(Vector skullStrippers)
  {
    // Find the maximum intensity
    int maxI = 0;
    short[][] array = null;
    BufferedImage image = null;
    SkullStripper ac = null;
    for(int z = 0; z < skullStrippers.size(); z++ ){
      ac = (SkullStripper)skullStrippers.elementAt(z);
      image = ac.getInputImage();
      array = ac.create2DArray(image);
      for(int x = 0; x < array.length; x++ ){
	for( int y = 0; y < array[0].length; y++ ){
	  if(maxI < array[x][y]) maxI = array[x][y];
	}
      }
    }

    // Initialize the histogram array
    int[] hist = new int[maxI + 1];

    // Construct histogram array
    int intens;
    for(int z = 0; z < skullStrippers.size(); z++ ){
      ac = (SkullStripper)skullStrippers.elementAt(z);
      image = ac.getInputImage();
      array = ac.create2DArray(image);
      for(int x = 0; x < array.length; x++ ){
	for( int y = 0; y < array[0].length; y++ ){
	  intens = array[x][y];
	  hist[intens]++;
	}
      }
    }

    // Calculate _intens2
    double sum = 0;
    ac = (SkullStripper)skullStrippers.elementAt(0);
    image = ac.getInputImage();
    int width = image.getWidth();
    int height = image.getHeight();
    double totalSum = width * height * skullStrippers.size();
    for(int k = 0; k < hist.length; k++){
	sum += hist[k];
	if(sum / totalSum > 0.02) {
	    _intens2 = k;
	    k = hist.length;
	}
    }

    // Calculate _intens98
    sum = 0;
    for(int k = 0; k < hist.length; k++){
	sum += hist[k];
	if(sum / totalSum > 0.98) {
	    _intens98 = k;
	    k = hist.length;
	}
    }

   // Calculate _intense10
    _intens10 = Math.floor(((_intens98 - _intens2)*0.1) + _intens2 + 0.5);

    // Calculate _intens40
    _intens40 = Math.floor(((_intens98 - _intens2)*0.4) + _intens2 + 0.5);
  }

  /**
   * If the segmentation results, represented by phi, is acceptable.
   *
   * @param phi to check on.
   * @param lastPhi Phi in last (or previous) slide.
   * @return True is phi is acceptable; or false if it is unacceptable.
   */
  private boolean _isPhiAcceptable(double[][] phi, SkullStripper ac, 
				   double[][] lastPhi)
  { 
    boolean isAcceptable = true;
    double LIMIT_JACCARD = 0.85;
    if(true){
      double[] temp = _evaluate(lastPhi, phi);
      double jaccard = temp[0];
      int maskArea = ac.calculateMaskArea(phi);
      if( maskArea > 10000){
	  if(jaccard < LIMIT_JACCARD) isAcceptable = false;
      }
      else {
	if(true){
	  int lastMaskArea = 0;
	  int contourLength = 0;
	  lastMaskArea = ac.calculateMaskArea(lastPhi);
	  double difference = maskArea - lastMaskArea;
      
	  double MAX_AREA_DIF = 0;
	  contourLength = ac.calculateContourLength(lastPhi);
	  if(contourLength > 0) 
	      MAX_AREA_DIF = contourLength *(Math.abs(INTER_SLICE_DIST)+1) * 2;
	  else MAX_AREA_DIF = difference; 
	  
	  // If the difference between lastMaskArea and maskArea is within 
	  // 15% of lastMaskArea, then phi is acceptable
	  if( difference > MAX_AREA_DIF ) isAcceptable = false;
	}
      }
    }
    return isAcceptable;
  }
}
